@@ -5,11 +5,8 @@ import platform
 from typing import Any
 from typing import Callable
 from typing import cast
-from typing import Dict
 from typing import IO
-from typing import List
 from typing import Optional
-from typing import Tuple
 
 from cryptography import x509
 from cryptography.hazmat.backends import default_backend
@@ -37,6 +34,7 @@ from certbot._internal import cli
 from certbot._internal import constants
 from certbot._internal import eff
 from certbot._internal import error_handler
+from certbot._internal import san
 from certbot._internal import storage
 from certbot._internal.plugins import disco as plugin_disco
 from certbot._internal.plugins import selection as plugin_selection
@@ -48,12 +46,14 @@ from certbot.interfaces import AccountStorage
 logger = logging.getLogger(__name__)
 
 
-def acme_from_config_key(config: configuration.NamespaceConfig,
-                         key: jose.JWK,
+def create_acme_client(config: configuration.NamespaceConfig,
+                         key: Optional[jose.JWK] = None,
                          regr: Optional[messages.RegistrationResource] = None,
+                         server_override: Optional[str] = None,
                          ) -> acme_client.ClientV2:
     """Wrangle ACME client construction"""
-    if key.typ == 'EC':
+    alg = RS256
+    if key and key.typ == 'EC':
         public_key = key.key
         if public_key.key_size == 256:
             alg = ES256
@@ -65,13 +65,14 @@ def acme_from_config_key(config: configuration.NamespaceConfig,
             raise errors.NotSupportedError(
                 "No matching signing algorithm can be found for the key"
             )
-    else:
-        alg = RS256
     net = acme_client.ClientNetwork(key, alg=alg, account=regr,
                                     verify_ssl=(not config.no_verify_ssl),
                                     user_agent=determine_user_agent(config))
 
-    directory = acme_client.ClientV2.get_directory(config.server, net)
+    server = config.server
+    if server_override:
+        server = server_override
+    directory = acme_client.ClientV2.get_directory(server, net)
     return acme_client.ClientV2(directory, net)
 
 
@@ -149,7 +150,7 @@ def sample_user_agent() -> str:
 
 def register(config: configuration.NamespaceConfig, account_storage: AccountStorage,
              tos_cb: Optional[Callable[[str], None]] = None
-             ) -> Tuple[account.Account, acme_client.ClientV2]:
+             ) -> tuple[account.Account, acme_client.ClientV2]:
     """Register new account with an ACME CA.
 
     This function takes care of generating fresh private key,
@@ -200,7 +201,7 @@ def register(config: configuration.NamespaceConfig, account_storage: AccountStor
             key_size=config.rsa_key_size,
             backend=default_backend())
     key = jose.JWKRSA(key=jose.ComparableRSAKey(rsa_key))
-    acme = acme_from_config_key(config, key)
+    acme = create_acme_client(config, key)
     # TODO: add phone?
     regr = perform_registration(acme, config, tos_cb)
 
@@ -225,15 +226,18 @@ def perform_registration(acme: acme_client.ClientV2, config: configuration.Names
     :returns: Registration Resource.
     :rtype: `acme.messages.RegistrationResource`
     """
+    if not acme.net.key:
+        raise errors.Error("acme client with no private key cannot register account.")
 
     eab_credentials_supplied = config.eab_kid and config.eab_hmac_key
-    eab: Optional[Dict[str, Any]]
+    eab: Optional[dict[str, Any]]
     if eab_credentials_supplied:
         account_public_key = acme.net.key.public_key()
         eab = messages.ExternalAccountBinding.from_data(account_public_key=account_public_key,
                                                         kid=config.eab_kid,
                                                         hmac_key=config.eab_hmac_key,
-                                                        directory=acme.directory)
+                                                        directory=acme.directory,
+                                                        hmac_alg=config.eab_hmac_alg)
     else:
         eab = None
 
@@ -290,7 +294,7 @@ class Client:
 
         # Initialize ACME if account is provided
         if acme is None and self.account is not None:
-            acme = acme_from_config_key(config, self.account.key, self.account.regr)
+            acme = create_acme_client(config, self.account.key, self.account.regr)
         self.acme = acme
 
         self.auth_handler: Optional[auth_handler.AuthHandler]
@@ -302,7 +306,7 @@ class Client:
 
     def obtain_certificate_from_csr(self, csr: util.CSR,
                                     orderr: Optional[messages.OrderResource] = None
-                                    ) -> Tuple[bytes, bytes]:
+                                    ) -> tuple[bytes, bytes]:
         """Obtain certificate.
 
         :param .util.CSR csr: PEM-encoded Certificate Signing
@@ -345,13 +349,13 @@ class Client:
         cert, chain = crypto_util.cert_and_chain_from_fullchain(fullchain)
         return cert.encode(), chain.encode()
 
-    def obtain_certificate(self, domains: List[str], old_keypath: Optional[str] = None
-                           ) -> Tuple[bytes, bytes, util.Key, util.CSR]:
+    def obtain_certificate(self, sans: list[san.SAN], old_keypath: Optional[str] = None
+                           ) -> tuple[bytes, bytes, util.Key, util.CSR]:
         """Obtains a certificate from the ACME server.
 
         `.register` must be called before `.obtain_certificate`
 
-        :param list domains: domains to get a certificate
+        :param list sans: domains and/or IP addresses for which to get a certificate.
 
         :returns: certificate as PEM string, chain as PEM string,
             newly generated private key (`.util.Key`), and DER-encoded
@@ -395,6 +399,9 @@ class Client:
         elif self.config.rsa_key_size and self.config.key_type.lower() == 'rsa':
             key_size = self.config.rsa_key_size
 
+        domains, _ = san.split(sans)
+        domains_str = [d.dns_name for d in domains]
+
         # Create CSR from names
         if self.config.dry_run:
             key = key or util.Key(
@@ -408,7 +415,8 @@ class Client:
             )
             csr = util.CSR(file=None, form="pem",
                            data=acme_crypto_util.make_csr(
-                               key.pem, domains, self.config.must_staple))
+                               key.pem, domains_str, self.config.must_staple,
+                               ))
         else:
             key = key or crypto_util.generate_key(
                 key_size=key_size,
@@ -418,43 +426,43 @@ class Client:
                 strict_permissions=self.config.strict_permissions,
             )
             csr = crypto_util.generate_csr(
-                key, domains, None, self.config.must_staple, self.config.strict_permissions)
+                key, domains_str, None, self.config.must_staple, self.config.strict_permissions)
 
         try:
             orderr = self._get_order_and_authorizations(csr.data, self.config.allow_subset_of_names)
         except messages.Error as error:
-            # Some domains may be rejected during order creation.
+            # Some sans may be rejected during order creation.
             # Certbot can retry the operation without the rejected
-            # domains contained within subproblems.
+            # sans contained within subproblems.
             if self.config.allow_subset_of_names:
-                successful_domains = self._successful_domains_from_error(error, domains)
-                if successful_domains != domains and len(successful_domains) != 0:
-                    return self._retry_obtain_certificate(domains, successful_domains, old_keypath)
+                successful_sans = self._successful_sans_from_error(error, sans)
+                if successful_sans != sans and len(successful_sans) != 0:
+                    return self._retry_obtain_certificate(sans, successful_sans, old_keypath)
             raise
         authzr = orderr.authorizations
-        auth_domains = {a.body.identifier.value for a in authzr}
-        successful_domains = [d for d in domains if d in auth_domains]
+        auth_ident_values = {a.body.identifier.value for a in authzr}
+        successful_sans = [s for s in sans if str(s) in auth_ident_values]
 
         # allow_subset_of_names is currently disabled for wildcard
         # certificates. The reason for this and checking allow_subset_of_names
-        # below is because successful_domains == domains is never true if
-        # domains contains a wildcard because the ACME spec forbids identifiers
+        # below is because successful_sans == sans is never true if
+        # sans contains a wildcard because the ACME spec forbids identifiers
         # in authzs from containing a wildcard character.
-        if self.config.allow_subset_of_names and successful_domains != domains:
-            return self._retry_obtain_certificate(domains, successful_domains, old_keypath)
+        if self.config.allow_subset_of_names and successful_sans != sans:
+            return self._retry_obtain_certificate(sans, successful_sans, old_keypath)
         else:
             try:
                 cert, chain = self.obtain_certificate_from_csr(csr, orderr)
                 return cert, chain, key, csr
             except messages.Error as error:
-                # Some domains may be rejected during the very late stage of
+                # Some sans may be rejected during the very late stage of
                 # order finalization. Certbot can retry the operation without
-                # the rejected domains contained within subproblems.
+                # the rejected sans contained within subproblems.
                 if self.config.allow_subset_of_names:
-                    successful_domains = self._successful_domains_from_error(error, domains)
-                    if successful_domains != domains and len(successful_domains) != 0:
+                    successful_sans = self._successful_sans_from_error(error, sans)
+                    if successful_sans != sans and len(successful_sans) != 0:
                         return self._retry_obtain_certificate(
-                            domains, successful_domains, old_keypath)
+                            sans, successful_sans, old_keypath)
                 raise
 
     def _get_order_and_authorizations(self, csr_pem: bytes,
@@ -471,8 +479,17 @@ class Client:
         """
         if not self.acme:
             raise errors.Error("ACME client is not set.")
+
+        profile = None
+        available_profiles = self.acme.directory.meta.profiles
+        preferred_profile = self.config.preferred_profile
+        if self.config.required_profile is not None:
+            profile = self.config.required_profile
+        elif (preferred_profile and available_profiles and
+              preferred_profile in available_profiles):
+            profile = preferred_profile
         try:
-            orderr = self.acme.new_order(csr_pem)
+            orderr = self.acme.new_order(csr_pem, profile=profile)
         except acme_errors.WildcardUnsupportedError:
             raise errors.Error("The currently selected ACME CA endpoint does"
                                " not support issuing wildcard certificates.")
@@ -485,7 +502,7 @@ class Client:
             deactivated, failed = self.auth_handler.deactivate_valid_authorizations(orderr)
             if deactivated:
                 logger.debug("Recreating order after authz deactivations")
-                orderr = self.acme.new_order(csr_pem)
+                orderr = self.acme.new_order(csr_pem, profile=profile)
             if failed:
                 logger.warning("Certbot was unable to obtain fresh authorizations for every domain"
                                ". The dry run will continue, but results may not be accurate.")
@@ -493,16 +510,16 @@ class Client:
         authzr = self.auth_handler.handle_authorizations(orderr, self.config, best_effort)
         return orderr.update(authorizations=authzr)
 
-    def obtain_and_enroll_certificate(self, domains: List[str], certname: Optional[str]
+    def obtain_and_enroll_certificate(self, sans: list[san.SAN], certname: Optional[str]
                                       ) -> Optional[storage.RenewableCert]:
         """Obtain and enroll certificate.
 
-        Get a new certificate for the specified domains using the specified
+        Get a new certificate for the specified domains and/or IP addresses using the specified
         authenticator and installer, and then create a new renewable lineage
         containing it.
 
-        :param domains: domains to request a certificate for
-        :type domains: `list` of `str`
+        :param sans: domains and/or IP addresses to request a certificate for
+        :type sans: `list` of `san.SAN`
         :param certname: requested name of lineage
         :type certname: `str` or `None`
 
@@ -510,8 +527,8 @@ class Client:
             referred to the enrolled cert lineage, or None if doing a successful dry run.
 
         """
-        new_name = self._choose_lineagename(domains, certname)
-        cert, chain, key, _ = self.obtain_certificate(domains)
+        new_name = self._choose_lineagename(sans, certname)
+        cert, chain, key, _ = self.obtain_certificate(sans)
 
         if (self.config.config_dir != constants.CLI_DEFAULTS["config_dir"] or
                 self.config.work_dir != constants.CLI_DEFAULTS["work_dir"]):
@@ -527,29 +544,38 @@ class Client:
             key.pem, chain,
             self.config)
 
-    def _successful_domains_from_error(self, error: messages.Error, domains: List[str],
-                                ) -> List[str]:
+    def _successful_sans_from_error(self, error: messages.Error, sans: list[san.SAN],
+                                    ) -> list[san.SAN]:
         if error.subproblems is not None:
-            failed_domains = [problem.identifier.value for problem in error.subproblems
-                                if problem.identifier is not None]
-            successful_domains = [x for x in domains if x not in failed_domains]
-            return successful_domains
+            failed_sans: list[san.SAN] = []
+            for problem in error.subproblems:
+                if not problem.identifier:
+                    continue
+                match problem.identifier.typ:
+                    case messages.IDENTIFIER_FQDN:
+                        failed_sans.append(san.DNSName(problem.identifier.value))
+                    case messages.IDENTIFIER_IP:
+                        failed_sans.append(san.IPAddress(problem.identifier.value))
+                    case _:
+                        raise TypeError(f"invalid identifier type {problem.identifier.typ}")
+            successful_sans = [x for x in sans if x not in failed_sans]
+            return successful_sans
         return []
 
-    def _retry_obtain_certificate(self, domains: List[str], successful_domains: List[str],
+    def _retry_obtain_certificate(self, sans: list[san.SAN], successful_sans: list[san.SAN],
                                 old_keypath: Optional[str]
-                                ) -> Tuple[bytes, bytes, util.Key, util.CSR]:
-        failed_domains = [d for d in domains if d not in successful_domains]
-        domains_list = ", ".join(failed_domains)
+                                ) -> tuple[bytes, bytes, util.Key, util.CSR]:
+        failed_sans = [s for s in sans if s not in successful_sans]
+        failed_sans_list = san.display(failed_sans)
         display_util.notify("Unable to obtain a certificate with every requested "
-            f"domain. Retrying without: {domains_list}")
-        return self.obtain_certificate(successful_domains, old_keypath)
+            f"domain. Retrying without: {failed_sans_list}")
+        return self.obtain_certificate(successful_sans, old_keypath)
 
-    def _choose_lineagename(self, domains: List[str], certname: Optional[str]) -> str:
+    def _choose_lineagename(self, sans: list[san.SAN], certname: Optional[str]) -> str:
         """Chooses a name for the new lineage.
 
-        :param domains: domains in certificate request
-        :type domains: `list` of `str`
+        :param sans: domains and/or IP addresses in certificate request
+        :type sans: `list` of `san.SAN`
         :param certname: requested name of lineage
         :type certname: `str` or `None`
 
@@ -563,11 +589,11 @@ class Client:
         lineagename = None
         if certname:
             lineagename = certname
-        elif util.is_wildcard_domain(domains[0]):
+        elif sans[0].is_wildcard():
             # Don't make files and directories starting with *.
-            lineagename = domains[0][2:]
+            lineagename = str(sans[0])[2:]
         else:
-            lineagename = domains[0]
+            lineagename = str(sans[0])
         # Verify whether chosen lineage is valid
         if self._is_valid_lineagename(lineagename):
             return lineagename
@@ -594,7 +620,7 @@ class Client:
 
     def save_certificate(self, cert_pem: bytes, chain_pem: bytes,
                          cert_path: str, chain_path: str, fullchain_path: str
-                         ) -> Tuple[str, str, str]:
+                         ) -> tuple[str, str, str]:
         """Saves the certificate received from the ACME server.
 
         :param bytes cert_pem:
@@ -629,11 +655,11 @@ class Client:
 
         return abs_cert_path, abs_chain_path, abs_fullchain_path
 
-    def deploy_certificate(self, domains: List[str], privkey_path: str, cert_path: str,
+    def deploy_certificate(self, sans: list[san.DNSName], privkey_path: str, cert_path: str,
                            chain_path: str, fullchain_path: str) -> None:
         """Install certificate
 
-        :param list domains: list of domains to install the certificate
+        :param list sans: list of domains/and or IP addresses to install the certificate
         :param str privkey_path: path to certificate private key
         :param str cert_path: certificate file path (optional)
         :param str fullchain_path: path to the full chain of the certificate
@@ -650,10 +676,13 @@ class Client:
         display_util.notify("Deploying certificate")
 
         msg = "Could not install certificate"
+        domains, ip_addresses = san.split(sans)
+        if ip_addresses:
+            raise TypeError("deploy of IP address certificate not supported")
         with error_handler.ErrorHandler(self._recovery_routine_with_msg, msg):
             for dom in domains:
                 self.installer.deploy_cert(
-                    domain=dom, cert_path=os.path.abspath(cert_path),
+                    domain=dom.dns_name, cert_path=os.path.abspath(cert_path),
                     key_path=os.path.abspath(privkey_path),
                     chain_path=chain_path,
                     fullchain_path=fullchain_path)
@@ -668,18 +697,17 @@ class Client:
             # sites may have been enabled / final cleanup
             self.installer.restart()
 
-    def enhance_config(self, domains: List[str], chain_path: str,
+    def enhance_config(self, domains: list[san.DNSName], chain_path: str,
                        redirect_default: bool = True) -> None:
         """Enhance the configuration.
 
-        :param list domains: list of domains to configure
+        :param list domains: list of domains to configure.
         :param chain_path: chain file path
         :type chain_path: `str` or `None`
         :param redirect_default: boolean value that the "redirect" flag should default to
 
         :raises .errors.Error: if no installer is specified in the
             client.
-
         """
         if self.installer is None:
             logger.error("No installer is specified, there isn't any "
@@ -712,11 +740,11 @@ class Client:
             with error_handler.ErrorHandler(self._rollback_and_restart, msg):
                 self.installer.restart()
 
-    def apply_enhancement(self, domains: List[str], enhancement: str,
+    def apply_enhancement(self, domains: list[san.DNSName], enhancement: str,
                           options: Optional[str] = None) -> None:
         """Applies an enhancement on all domains.
 
-        :param list domains: list of ssl_vhosts (as strings)
+        :param list domains: list of ssl_vhosts (as san.DNSName)
         :param str enhancement: name of enhancement, e.g. ensure-http-header
         :param str options: options to enhancement, e.g. Strict-Transport-Security
 
@@ -733,7 +761,7 @@ class Client:
         with error_handler.ErrorHandler(self._recovery_routine_with_msg, None):
             for dom in domains:
                 try:
-                    self.installer.enhance(dom, enhancement, options)
+                    self.installer.enhance(dom.dns_name, enhancement, options)
                 except errors.PluginEnhancementAlreadyPresent:
                     logger.info("Enhancement %s was already set.", enh_label)
                 except errors.PluginError:
@@ -844,7 +872,7 @@ def rollback(default_installer: str, checkpoints: int,
 
 
 def _open_pem_file(config: configuration.NamespaceConfig,
-                   cli_arg_path: str, pem_path: str) -> Tuple[IO, str]:
+                   cli_arg_path: str, pem_path: str) -> tuple[IO, str]:
     """Open a pem file.
 
     If cli_arg_path was set by the client, open that.
